import json
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict

from jsonschema import Draft7Validator
from jsonschema.exceptions import ValidationError
from yaml import safe_load

from pyinsights.exceptions import (
    ConfigInvalidSyntaxError,
    ConfigNotFoundError,
    ConfigVersionUnknownError,
    InvalidVersionError,
)
from pyinsights.helper import (
    convert_to_epoch,
    convert_string_duration_to_datetime,
)


ConfigType = Dict[str, Any]
SchemaType = Dict[str, Any]


@dataclass
class ConfigFile:
    filename: str
    content: ConfigType

    @classmethod
    def from_filename(cls, filename: str) -> "ConfigFile":
        return cls(filename, load_yaml(filename))

    @property
    def version(self) -> str:
        try:
            return self.content["version"]
        except KeyError:
            raise ConfigVersionUnknownError(
                "Please Specify configuration version"
            )

    def convert_duration(self) -> Dict[str, int]:
        duration = self.content["duration"]

        if isinstance(duration, str):
            duration = convert_string_duration_to_datetime(duration)

        duration_epoch = {
            key: convert_to_epoch(value) for key, value in duration.items()
        }
        return duration_epoch

    def get_query_params(self) -> ConfigType:
        params = self.content.copy()
        new_duration = self.convert_duration()
        del params["version"]
        del params["duration"]
        params.update(new_duration)
        return params


def load_config(filepath: str) -> ConfigFile:
    """Load configuration

    Arguments:
        filepath {str}

    Returns:
        {ConfigFile} -- query parameters
    """

    config = ConfigFile.from_filename(filepath)
    validate(config.content, config.version)
    return config


def load_yaml(filepath: str) -> ConfigType:
    """Load YAML configuration file

    Arguments:
        filepath {str}

    Raises:
        ConfigNotFoundError

    Returns:
        config {ConfigType}
    """

    try:
        with open(filepath) as fobj:
            return safe_load(fobj)
    except FileNotFoundError:
        raise ConfigNotFoundError("Could not find the configuration")


def load_schema(version: str) -> SchemaType:
    """Load the schema json file

    Arguments:
        version {str}

    Raises:
        InvalidVersionError

    Returns:
        schema {SchemaType}
    """

    basepath = Path(__file__).parent.resolve()
    filename = f"version_{version}.json"
    schema_filpath = f"{basepath}/schema/{filename}"

    try:
        with open(schema_filpath) as fobj:
            return json.load(fobj)
    except FileNotFoundError:
        raise InvalidVersionError(f"The version {repr(version)} is invalid")


def validate(config: ConfigType, version: str) -> bool:
    """Validate the configuration

    Arguments:
        config {ConfigType}
        version {str}

    Raises:
        ConfigInvalidSyntaxError

    Returns:
        bool
    """

    try:
        schema = load_schema(version)
        Draft7Validator(schema).validate(config)
    except ValidationError as err:
        raise ConfigInvalidSyntaxError(err)
    except Exception as err:
        raise err
    else:
        return True



import json
import shutil
from typing import Dict, List


QueryResult = List[List[Dict[str, str]]]


class Formatter:
    def __init__(self, format_type: str = "json") -> None:
        """
        Keyword Arguments:
            format_type {str} -- [description] (default: {'json'})
        """
        self.format_type = format_type

    @staticmethod
    def _to_pretty_dict(results: QueryResult) -> List[Dict[str, str]]:
        """Format results to python dict in list

        Arguments:
            results {QueryResult}

        Returns:
            List[Dict[str, str]]
        """

        formatted_result = [
            {
                field["field"]: field["value"]
                for field in result
                if field["field"] != "@ptr"
            }
            for result in results
        ]
        return formatted_result

    def to_json(self, results: QueryResult) -> str:
        """Format results to json

        Arguments:
            results {QueryResult}

        Returns:
            str
        """

        if not results:
            return ""

        tmp_resuls = self._to_pretty_dict(results)
        formatted_result = json.dumps(tmp_resuls, indent=2, ensure_ascii=False)
        return formatted_result

    def to_table(self, results: QueryResult) -> str:
        """Format results to string table

        Arguments:
            results {QueryResult}

        Returns:
            str
        """

        if not results:
            return ""

        tmp_results = self._to_pretty_dict(results)
        headers = list(tmp_results[0].keys())
        width, _ = shutil.get_terminal_size()
        length_per_column = width // len(headers)

        table_header = ""
        for header in headers:
            table_header += header.ljust(length_per_column)

        table_record = ""
        for result in tmp_results:
            for field in result.values():
                table_record += field[: length_per_column - 2].ljust(
                    length_per_column
                )
            table_record += "\n"

        formatted_result = f"{table_header}\n{table_record}"
        return formatted_result


def format_result(format_type: str, results: QueryResult) -> str:
    """Format the query result

    Arguments:
        format_type {str}: json or table
        results {QueryResult}

    Returns:
        str
    """

    formatter = Formatter(format_type)
    func = getattr(formatter, f"to_{format_type}")
    formatted_result = func(results)
    return formatted_result



import argparse
from typing import Any, Dict
import sys

from pyinsights.__version__ import __version__
from pyinsights.config import load_config
from pyinsights.query import query
from pyinsights.formatter import format_result


CliOptions = Dict[str, Any]


def parse_args() -> Dict[str, Any]:
    """Parse arguments

    Returns:
        Dict[str, Any]
    """

    parser = argparse.ArgumentParser(
        prog="pyinsights",
        description="AWS CloudWatch Logs Insights is wrapped by Python",
    )

    parser.add_argument(
        "-c",
        "--config",
        required=True,
        default="pyinsights.yml",
        help="PyInsights config file path",
    )

    parser.add_argument(
        "-f",
        "--format",
        choices=["json", "table"],
        default="json",
        help='Output format "json" or "table"',
    )

    parser.add_argument("-p", "--profile", help="AWS profile name")

    parser.add_argument("-r", "--region", help="AWS region")

    parser.add_argument(
        "-v", "--version", action="version", version=__version__
    )

    return vars(parser.parse_args())


def run(cli_options: CliOptions) -> bool:
    config = load_config(cli_options["config"])
    tmp_result = query(
        cli_options["region"],
        cli_options["profile"],
        config.get_query_params(),
    )

    if isinstance(tmp_result, dict) and (results := tmp_result.get("results")):
        formatted_result = format_result(cli_options["format"], results)  # type: ignore
        sys.stdout.write(formatted_result)
        return True

    return False


def main() -> bool:
    args = parse_args()
    sys.exit(run(args))



import sys
from random import randint
from datetime import datetime, timedelta
from typing import Dict, Union

from pyinsights.exceptions import InvalidDurationError


def convert_to_epoch(duration: Union[str, datetime]) -> int:
    """Convert datetime string to epoch (POSIX timestamp)

    Arguments:
        duration {Union[str, datetime]}
            --  string format must be `%Y-%m-%d %H:%M:%S`
                if duration type is str

    Raises:
        InvalidDurationError

    Returns:
        epoch {int}
    """

    if isinstance(duration, str):
        time_format = "%Y-%m-%d %H:%M:%S"
        try:
            duration = datetime.strptime(duration, time_format)
        except ValueError:
            raise InvalidDurationError(
                f"{duration=} is invalid datetime format as \
                    duration parameter"
            )

    if not isinstance(duration, datetime):
        raise InvalidDurationError(
            f"Cloud not convert {duration=} to POSIX timestamp"
        )

    epoch = int(duration.timestamp())
    return epoch


TIME_UNITS = {
    "s": "seconds",
    "m": "minutes",
    "h": "hours",
    "d": "days",
    "w": "weeks",
}


def convert_string_duration_to_datetime(
    string_duration: str,
) -> Dict[str, datetime]:
    """Convert string duration to datetime

    Arguments:
        string_duration {str}

    Raises:
        InvalidDurationError

    Returns:
        Dict[str, datetime] -- `start_time` and `end_time` are key
    """

    try:
        duration = {TIME_UNITS[string_duration[-1]]: int(string_duration[:-1])}
    except (ValueError, IndexError, KeyError):
        raise InvalidDurationError(
            f"{string_duration=} is invalid as duration parameter"
        )

    end_time = datetime.now()
    start_time = end_time - timedelta(**duration)
    duraion_map = {"start_time": start_time, "end_time": end_time}
    return duraion_map


class Color:
    @classmethod
    def ansi(cls, code: Union[str, int]) -> str:
        """Get ansi

        Arguments:
            code {Union[str, int]}

        Returns:
            str
        """

        return f"\033[{code}m"

    @classmethod
    def select_color_randomly(cls) -> str:
        """Select a color randomly

        Returns:
            str
        """

        code = randint(1, 6)
        return cls.ansi(30 + code)

    @classmethod
    def disabled(cls) -> str:
        """Disable color

        Returns:
            str
        """

        return cls.ansi("0")

    @classmethod
    def bold(cls) -> str:
        """Get bold code

        Returns:
            str
        """

        return cls.ansi("01")


def processing(msg: str, end: str = "") -> None:
    """Display processing on terminal

    Arguments:
        msg {str}

    Keyword Arguments:
        end {str} -- (default: {''})
    """

    processing_msg = f"{Color.bold()}{Color.select_color_randomly()}{msg}{Color.disabled()}{end}"
    sys.stdout.write(processing_msg)
    sys.stdout.flush()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyinsights']

package_data = \
{'': ['*'], 'pyinsights': ['schema/*']}

install_requires = \
['boto3>=1.10.45,<2.0.0', 'jsonschema>=3.2.0,<4.0.0', 'pyyaml>=5.2,<6.0']

entry_points = \
{'console_scripts': ['pyinsights = pyinsights.cli:main']}

setup_kwargs = {
    'name': 'pyinsights',
    'version': '0.3.1',
    'description': 'AWS CloudWatch Logs Insights is wrapped by Python',
    'long_description': "# PyInsights\n\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/pyinsights)\n![PyPI](https://img.shields.io/pypi/v/pyinsights?color=blue)\n![GitHub](https://img.shields.io/github/license/homoluctus/pyinsights)\n\nA CLI tool To query CloudWatch Logs Insights.\n\n![usage1](https://raw.githubusercontent.com/homoluctus/pyinsights/master/images/usage1.png)\n\n![usage2](https://raw.githubusercontent.com/homoluctus/pyinsights/master/images/usage2.png)\n\n# ToC\n\n<!-- TOC depthFrom:2 -->\n\n- [Usage](#usage)\n  - [Write Configuration](#write-configuration)\n  - [Execute command](#execute-command)\n- [Configuration](#configuration)\n- [CLI Options](#cli-options)\n- [Environment Variable](#environment-variable)\n\n<!-- /TOC -->\n\n## Usage\n\n### Write Configuration\n\nWrite configuration to `pyinsights.yml` like:\n\n```yaml\nversion: '1.0'\nlog_group_name:\n  - '/ecs/sample'\nquery_string: 'field @message | filter @message like /ERROR/'\nduration: '30m'\nlimit: 10\n```\n\nI wrote examples, so see [examples folder](https://github.com/homoluctus/pyinsights/tree/master/examples).\n\n### Execute command\n\n```bash\npyinsights -c pyinsights.yml -p aws_profile -r region\n```\n\n## Configuration\n\n|Parameter|Type|Required|Description|\n|:--:|:--:|:--:|:--|\n|version|string|true|Choose configuration version from ['1.0']|\n|log_group_name|array|true|Target log group names to query|\n|query_string|string|true|Pattern to query|\n|duration|string or object|true||\n||string||Specify hours, minutes or seconds from now<br>Unit:<br>hours = `h`,<br>minutes = `m`,<br>seconds = `s`,<br>days = `d`,<br>weeks = `w`|\n||object||Specify `start_time` and `end_time`<br>Datetime format must be `YYYY-MM-DD HH:MM:SS`|\n|limit|integer|false|The number of log to fetch|\n\n## CLI Options\n\n|Option|Required|Description|\n|:--:|:--:|:--|\n|-c, --config|true|Specify yaml configuration by absolute or relative path|\n|-f, --format|false|Choose from json or table|\n|-p, --profile|false|AWS profile name|\n|-r, --region|false|AWS region|\n|-v, --version|false|Show version|\n\n## Environment Variable\n\nIf `profile` and `region` options are not specified, AWS Credentials must be set as environment variables.\n\n- AWS_ACCESS_KEY_ID\n- AWS_SECRET_ACCESS_KEY\n- AWS_DEFAULT_REGION\n\nPlease see [Environment Variable Configuration](https://boto3.amazonaws.com/v1/documentation/api/latest/guide/configuration.html#environment-variable-configuration) for the detail.\n",
    'author': 'homoluctus',
    'author_email': 'w.slife18sy@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/homoluctus/pyinsights',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# pylint: disable=R0913

import os
import sys
from time import sleep
import concurrent.futures as confu
from concurrent.futures.thread import ThreadPoolExecutor
from typing import Any, Dict, List, Optional

import boto3
import botocore.errorfactory

from pyinsights.config import ConfigType
from pyinsights.exceptions import (
    NotFetchQueryResultError,
    QueryAlreadyCancelled,
    QueryNotYetStartError,
    QueryTimeoutError,
)
from pyinsights.helper import processing


AWS_ACCESS_KEY_ID = os.getenv("AWS_ACCESS_KEY_ID")
AWS_SECRET_ACCESS_KEY = os.getenv("AWS_SECRET_ACCESS_KEY")
AWS_DEFAULT_REGION = os.getenv("AWS_DEFAULT_REGION")


QueryResultResponse = Dict[str, Any]


class InsightsClient:
    def __init__(
        self, region: Optional[str] = None, profile: Optional[str] = None,
    ) -> None:
        """
        Keyword Arguments:
            region {Optional[str]}
            profile {Optional[str]}
        """

        region_name = region or AWS_DEFAULT_REGION

        session = boto3.Session(
            aws_access_key_id=AWS_ACCESS_KEY_ID,
            aws_secret_access_key=AWS_SECRET_ACCESS_KEY,
            region_name=region_name,
            profile_name=profile,
        )

        self.__client = session.client("logs")
        self.__query_id = None

    def start_query(
        self,
        query_string: str,
        start_time: int,
        end_time: int,
        log_group_name: List[str],
        limit: Optional[int] = 1000,
    ) -> bool:
        """Start query

        Arguments:
            query_string {str} -- query string
            start_time {int} -- POSIX timestamp
            end_time {int} -- POSIX timestamp
            log_group_name {List[str]} -- aws cloudwatch log group to query

        Keyword Arguments:
            limit {Optional[int] -- the maximum number of log events
                                    to return in the query. (default: {1000})

        Returns:
            bool
        """

        response = self.__client.start_query(
            logGroupNames=log_group_name,
            startTime=start_time,
            endTime=end_time,
            queryString=query_string,
            limit=limit,
        )

        self.__query_id = response["queryId"]
        return True

    def fetch_result(self, wait_time: float = 0.2) -> QueryResultResponse:
        """Fetch the query result

        Keyword Arguments:
            wait_time {float}
                -- wait time when query status is `Scheduled` or `Running`.

        Raises:
            QueryNotYetStartError

        Returns:
            results {QueryResultResponse}
        """

        if self.__query_id is None:
            raise QueryNotYetStartError("The Query has not yet started")

        results = self.__client.get_query_results(queryId=self.__query_id)
        status = results["status"]

        if status in ["Scheduled", "Running"]:
            sleep(wait_time)
            results.update(self.fetch_result())

        elif status == "Failed":
            raise NotFetchQueryResultError("Could not fetch the query result.")

        elif status == "Timeout":
            raise QueryTimeoutError("The query is timeout.")

        elif status == "Cancelled":
            raise QueryAlreadyCancelled(
                "The query has already been cancelled."
            )

        return results

    def end_query(self) -> bool:
        """Gracefully terminate query

        Raises:
            err:
                raise exception if InvalidParameterException,
                ServiceUnavailableException or other error occurs

        Returns:
            bool
        """

        if self.__query_id is None:
            return True

        try:
            result = self.__client.stop_query(queryId=self.__query_id)
            return result["success"]
        except botocore.errorfactory.ClientError as err:
            already_stopped_msg = "Query is not in Running or Scheduled state"
            if (
                err.operation_name == "StopQuery"
                and already_stopped_msg in err.__str__()
            ):
                return True
            raise err
        except Exception as err:
            raise err


def query(
    region: str, profile: str, query_params: ConfigType
) -> QueryResultResponse:
    """Run thread

    Arguments:
        region {str}
        profile {str}
        query_params {ConfigType}

    Returns:
        results = QueryResultResponse
    """

    results = {}
    processing("Waiting", end=" ")

    client = InsightsClient(region, profile)
    client.start_query(**query_params)

    try:
        with ThreadPoolExecutor(max_workers=1) as executor:
            thread = executor.submit(client.fetch_result)

            while True:
                try:
                    results = thread.result(timeout=0.5)
                except confu.TimeoutError:
                    processing(".", end="")
                else:
                    if results:
                        processing(".", end="\n")
                        break
    except KeyboardInterrupt:
        client.end_query()
        sys.exit("\nAbort")
    else:
        return results
